/*

	ent-ghost
	Copyright [2011-2013] [Jack Lu1]

	This file is part of the ent-ghost source code.

	ent-ghost is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	ent-ghost source code is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with ent-ghost source code. If not, see <http://www.gnu.org/licenses/>.

	ent-ghost is modified from GHost++ (http://ghostplusplus.googlecode.com/)
	GHost++ is Copyright [2008] [Trevor Hogan]

*/

#include "ghost.h"
#include "util.h"
#include "keyprotocol.h"

CKeyProtocol :: CKeyProtocol( )
{

}

CKeyProtocol :: ~CKeyProtocol( )
{

}

///////////////////////
// RECEIVE FUNCTIONS //
///////////////////////

CKeyInfo *CKeyProtocol :: RECEIVE_KEY_OBTAIN( BYTEARRAY data )
{
	// 2 bytes					-> Length
	// 1 byte					-> ID
	// (STRING)					-> Server
	// (STRING)					-> Type
	// (STRING)					-> Key

	if( ValidateLength( data ) && data.size( ) >= 4 )
	{
		BYTEARRAY Server = UTIL_ExtractCString( data, 3 );

		if( !Server.empty( ) && data.size( ) >= Server.size( ) + 4 )
		{
			BYTEARRAY Type = UTIL_ExtractCString( data, 4 + Server.size( ) );

			if( !Type.empty( ) && data.size( ) >= Server.size( ) + Type.size( ) + 5 )
			{
				BYTEARRAY Key = UTIL_ExtractCString( data, 5 + Server.size( ) + Type.size( ) );
				return new CKeyInfo( string( Server.begin( ), Server.end( ) ), string( Type.begin( ), Type.end( ) ), string( Key.begin( ), Key.end( ) ) );
			}
		}
	}

	return NULL;
}

////////////////////
// SEND FUNCTIONS //
////////////////////

BYTEARRAY CKeyProtocol :: SEND_KEY_NULL( )
{
	BYTEARRAY packet;
	packet.push_back( 0 );							// packet length will be assigned later
	packet.push_back( 0 );							// packet length will be assigned later
	packet.push_back( KEY_NULL );					// BNLS_NULL
	AssignLength( packet );
	return packet;
}

BYTEARRAY CKeyProtocol :: SEND_KEY_AUTH( string token )
{
	BYTEARRAY packet;
	packet.push_back( 0 );								// packet length will be assigned later
	packet.push_back( 0 );								// packet length will be assigned later
	packet.push_back( KEY_AUTH );						// KEY_AUTH
	UTIL_AppendByteArrayFast( packet, token );			// authentication token
	AssignLength( packet );
	return packet;
}

BYTEARRAY CKeyProtocol :: SEND_KEY_OBTAIN( string server, string type )
{
	BYTEARRAY packet;
	packet.push_back( 0 );								// packet length will be assigned later
	packet.push_back( 0 );								// packet length will be assigned later
	packet.push_back( KEY_OBTAIN );						// KEY_AUTH
	UTIL_AppendByteArrayFast( packet, server );			// server
	UTIL_AppendByteArrayFast( packet, type );			// type
	AssignLength( packet );
	return packet;
}

/////////////////////
// OTHER FUNCTIONS //
/////////////////////

bool CKeyProtocol :: AssignLength( BYTEARRAY &content )
{
	// insert the actual length of the content array into bytes 1 and 2 (indices 0 and 1)

	BYTEARRAY LengthBytes;

	if( content.size( ) >= 2 && content.size( ) <= 65535 )
	{
		LengthBytes = UTIL_CreateByteArray( (uint16_t)content.size( ), true );
		content[0] = LengthBytes[0];
		content[1] = LengthBytes[1];
		return true;
	}

	return false;
}

bool CKeyProtocol :: ValidateLength( BYTEARRAY &content )
{
	// verify that bytes 1 and 2 (indices 0 and 1) of the content array describe the length

	uint16_t Length;
	BYTEARRAY LengthBytes;

	if( content.size( ) >= 2 && content.size( ) <= 65535 )
	{
		LengthBytes.push_back( content[0] );
		LengthBytes.push_back( content[1] );
		Length = UTIL_ByteArrayToUInt16( LengthBytes, true );

		if( Length == content.size( ) )
			return true;
	}

	return false;
}

//
// CKeyInfo
//

CKeyInfo :: CKeyInfo( string nServer, string nType, string nKey ) : m_Server( nServer ), m_Type( nType ), m_Key( nKey )
{

}

CKeyInfo :: ~CKeyInfo( )
{

}
