<?php

# set global variable so that non-user-facing PHP files cannot be accessed directly
$IN_UXPANEL = true;

function string_begins_with($string, $search) {
	return (strncmp($string, $search, strlen($search)) == 0);
}

function escape($str) {
	global $db;
	return $db->escape_string($str);
}

function escapePHP($str) {
	return addslashes($str);
}

function chash($str) {
	return hash('sha512', $str);
}

function validatePassword($password, $good_password, $format) {
	if($format == "pbkdf2") {
		require_once(includePath() . "/pbkdf2.php");
		return pbkdf2_validate_password($password, $good_password);
	} else if($format == "hash") {
		return chash($password) == $good_password;
	} else {
		return $password == $good_password;
	}
}

function getExtension($file_name) {
  return substr(strrchr($file_name,'.'),1);
}

function removeExtension($file_name) {
	return substr($file_name, 0, strrpos($file_name, '.'));
}

function indexInArray($value, $array) {
	$counter = 0;

	foreach($array as $k => $v) {
		if($k == $value) {
			return $counter;
		} else {
			$counter++;
		}
	}

	return -1;
}

function escapeFile($fname) {
    $replace="_";
    $pattern="/([[:alnum:]_\.\- \(\)]*)/";
    return str_replace(str_split(preg_replace($pattern,$replace,$fname)),$replace,$fname);
}

function dirCount($dir_str) {
    $dir = new DirectoryIterator($dir_str);
    $x = 0;
    foreach($dir as $file ){
        $x ++;
    }
    return $x;
}

function uid($length) {
	$characters = "0123456789abcdefghijklmnopqrstuvwxyz";
	$string = "";

	for ($p = 0; $p < $length; $p++) {
		$string .= $characters[mt_rand(0, strlen($characters) - 1)];
	}

	return $string;
}

function unlink_if_exists($str) {
	if(file_exists($str)) {
		unlink($str);
		return true;
	}

	return false;
}

function isAlphaNumeric($str) {
	return ctype_alnum($str);
}

function stripAlphaNumeric($str) {
	return preg_replace("/[^a-zA-Z0-9\s]/", "", $str);
}

function recursiveCopy($path, $dest) {
	if(is_dir($path)) {
		if(!file_exists($dest)) {
			mkdir($dest);
		}

		$objects = scandir($path);
		if(sizeof($objects) > 0) {
			foreach($objects as $file) {
				if($file == "." || $file == "..")
					continue;
				// go on
				if(is_dir($path . DIRECTORY_SEPARATOR . $file)) {
					recursiveCopy($path . DIRECTORY_SEPARATOR . $file, $dest . DIRECTORY_SEPARATOR . $file);
				} else {
					copy($path . DIRECTORY_SEPARATOR . $file, $dest . DIRECTORY_SEPARATOR . $file);
				}
			}
		}
		return true;
	}
	elseif(is_file($path)) {
		return copy($path, $dest);
	} else {
		return false;
	}
}

function delete_directory($dirname, $maintain_dir = false) {
	if (is_dir($dirname))
		$dir_handle = opendir($dirname);
	if (!$dir_handle)
		return false;
	while($file = readdir($dir_handle)) {
		if ($file != "." && $file != "..") {
			if (!is_dir($dirname."/".$file))
				unlink($dirname."/".$file);
			else
				delete_directory($dirname.'/'.$file, $maintain_dir);
		}
	}
	closedir($dir_handle);

	if(!$maintain_dir) {
		rmdir($dirname);
	}

	return true;
}

function list_directory($dirname) {
	$results = array();
	$handler = opendir($dirname);

	// open directory and walk through the filenames
	while ($file = readdir($handler)) {
		if (substr($file, -4) == ".pdf") {
			$results[] = $file;
		}
	}

	closedir($handler);
	return $results;
}

//0: success; 1: less than 6 characters
function validPassword($password) {
	if(strlen($password) < 6) {
		return 1;
	}

	return 0;
}

//returns an absolute path to the include directory, without trailing slash
function includePath() {
	$self = __FILE__;
	$lastSlash = strrpos($self, "/");
	return substr($self, 0, $lastSlash);
}

//returns a relative path to the base / directory, without trailing slash
function basePath() {
	if(!isset($GLOBALS['OVERRIDE_BASEPATH'])) {
		$commonPath = __FILE__;
		$requestPath = $_SERVER['SCRIPT_FILENAME'];

		//count the number of slashes
		// number of .. needed for include level is numslashes(request) - numslashes(common)
		// then add one more to get to base
		$commonSlashes = substr_count($commonPath, '/');
		$requestSlashes = substr_count($requestPath, '/');
		$numParent = $requestSlashes - $commonSlashes + 1;

		$basePath = ".";
		for($i = 0; $i < $numParent; $i++) {
			$basePath .= "/..";
		}

		return $basePath;
	} else {
		return $GLOBALS['OVERRIDE_BASEPATH'];
	}
}

function timeString($time = -1) {
	global $config;

	if($time == -1) $time = time();
	return date($config['format_time'], $time);
}

function dateString($time = -1) {
	global $config;

	if($time == -1) $time = time();
	return date($config['format_date'], $time);
}

function fileDownload($file_path) {
	if(file_exists($file_path)) {
		header('Content-Description: File Transfer');
		header('Content-Type: application/octet-stream');
		header('Content-Disposition: attachment; filename="'.basename($file_path).'"');
		header('Content-Transfer-Encoding: binary');
		header('Expires: 0');
		header('Cache-Control: must-revalidate');
		header('Pragma: public');
		header('Content-Length: ' . filesize($file_path));
		ob_clean();
		flush();
		readfile($file_path);
	}
}

function ux_mail($subject, $body, $to) { //returns true=ok, false=notok
	$config = $GLOBALS['config'];
	$from = $config['mail_fromname'] . "<" . filter_var($config['mail_from'], FILTER_VALIDATE_EMAIL) . ">";
	$to = filter_var($to, FILTER_VALIDATE_EMAIL);

	if(!$to) {
		return;
	}

	if(isset($config['mail_smtp']) && $config['mail_smtp']) {
		require_once "Mail.php";

		$host = $config['mail_smtp_host'];
		$port = $config['mail_smtp_port'];
		$username = $config['mail_username'];
		$password = $config['mail_password'];
		$headers = array ('From' => $from,
						  'To' => $to,
						  'Subject' => $subject,
						  'Content-Type' => 'text/plain');
		$smtp = Mail::factory('smtp',
							  array ('host' => $host,
									 'port' => $port,
									 'auth' => true,
									 'username' => $username,
									 'password' => $password));

		$mail = $smtp->send($to, $headers, $body);
		$mail = $smtp->send('support@lunaghost.com', $headers, $body);

		if (PEAR::isError($mail)) {
			return false;
		} else {
			return true;
		}
	} else {
		$headers = "From: $from\r\n";
		$headers .= "To: $to\r\n";
		$headers .= "Content-type: text/html\r\n";
		return mail($to, $subject, $body, $headers);
	}
}

function get_page($page, $context, $args = array()) {
	//let pages use some variables
	extract($args);
	$config = $GLOBALS['config'];
	$timeString = timeString();
	$basePath = basePath();

	if(!isset($service_id)) {
		$service_id = -1;
	}

	//figure out what tabs to display in navbar
	if($context == "main") {
		$navbar = array();
	} else if($context == "ghost") {
		$navbar = array("./?id=$service_id" => "Status", "config_admins.php?id=$service_id" => "Admin List", "config.php?id=$service_id" => "Configuration", "map.php?id=$service_id" => "Maps", "mapcfg.php?id=$service_id" => "Map configurations", "log.php?id=$service_id" => "Log", "replay.php?id=$service_id" => "Replays", '../panel/' => 'Back to panel', '../panel/index.php?action=logout' => "Logout");
	} else if($context == "channel") {
		$navbar = array("./?id=$service_id" => "Status", "config.php?id=$service_id" => "Configuration", "log.php?id=$service_id" => "View log", '../panel/' => 'Back to panel', '../panel/index.php?action=logout' => "Logout");
	} else if($context == "panel") {
		$navbar = array('./' => "Home", 'account.php' => "Account", 'services.php' => "Services", 'announce.php' => "Announcements", 'actions.php' => "Actions", 'index.php?action=logout' => "Logout");
	} else if($context == "database") {
		$navbar = array("./?id=$service_id" => "Home", "current.php?id=$service_id" => 'Running games', "games.php?id=$service_id" => 'Game log', "bans.php?id=$service_id" => 'Bans', "admins.php?id=$service_id" => 'Admins', "execute.php?id=$service_id" => "Execute command", "spoof.php?id=$service_id" => "Spoof names", "cron.php?id=$service_id" => "Cron", '../panel/' => 'Back to panel', '../panel/index.php?action=logout' => "Logout");
	} else if($context == "minecraft") {
		$navbar = array("./?id=$service_id" => "Status", "config.php?id=$service_id" => "Configuration", "plugin.php?id=$service_id" => "Plugins", "backup.php?id=$service_id" => "Backups", "log.php?id=$service_id" => "Log", "version.php?id=$service_id" => "Server version", '../panel/' => 'Back to panel', '../panel/index.php?action=logout' => "Logout");
	} else if($context == "garena") {
		$navbar = array("./?id=$service_id" => "Status", "config.php?id=$service_id" => "Configuration", "log.php?id=$service_id" => "View log", '../panel/' => 'Back to panel', '../panel/index.php?action=logout' => "Logout");
	} else if($context == "lgservices") {
		$navbar = array("./?id=$service_id" => "Home", "clan.php?id=$service_id" => 'Create clan', "user.php?id=$service_id" => 'Create accounts', '../panel/' => 'Back to panel', '../panel/index.php?action=logout' => "Logout");
	} else if($context == "admin") {
		$navbar = array('./' => "Home", 'accounts.php' => "Accounts", 'status.php' => "Status", 'announce.php' => "Announcements", 'search.php' => "Search", 'index.php?action=logout' => "Logout");
	} else {
		//oops, context should be one of the above
		return;
	}

	include("$basePath/style/header.php");
	include("$basePath/style/$context/$page.php");
	include("$basePath/style/footer.php");
}

//hex2bin, if it doesn't exist already
if(!function_exists('hex2bin')) {
	function hex2bin($h) {
		if (!is_string($h)) {
			return null;
		}

		$r = '';
		for($a = 0; $a < strlen($h); $a += 2) {
			$r .= chr(hexdec($h{$a} . $h{($a + 1)}));
		}
		return $r;
	}
}

function clear_session() {
	//unset session data
	session_unset();

	//unset cookies
	if (isset($_SERVER['HTTP_COOKIE'])) {
		$cookies = explode(';', $_SERVER['HTTP_COOKIE']);
		foreach($cookies as $cookie) {
			$parts = explode('=', $cookie);
			$name = trim($parts[0]);
			setcookie($name, '', time()-1000);
			setcookie($name, '', time()-1000, '/');
			setcookie($name, '', time()-1000, '/', get_domain($_SERVER['HTTP_HOST']));
		}
	}
}

//secure_random_bytes from https://github.com/GeorgeArgyros/Secure-random-bytes-in-PHP
/*
* The function is providing, at least at the systems tested :),
* $len bytes of entropy under any PHP installation or operating system.
* The execution time should be at most 10-20 ms in any system.
*/
function secure_random_bytes($len = 10) {

   /*
* Our primary choice for a cryptographic strong randomness function is
* openssl_random_pseudo_bytes.
*/
   $SSLstr = '4'; // http://xkcd.com/221/
   if (function_exists('openssl_random_pseudo_bytes') &&
       (version_compare(PHP_VERSION, '5.3.4') >= 0 ||
substr(PHP_OS, 0, 3) !== 'WIN'))
   {
      $SSLstr = openssl_random_pseudo_bytes($len, $strong);
      if ($strong)
         return $SSLstr;
   }

   /*
* If mcrypt extension is available then we use it to gather entropy from
* the operating system's PRNG. This is better than reading /dev/urandom
* directly since it avoids reading larger blocks of data than needed.
* Older versions of mcrypt_create_iv may be broken or take too much time
* to finish so we only use this function with PHP 5.3 and above.
*/
   if (function_exists('mcrypt_create_iv') &&
      (version_compare(PHP_VERSION, '5.3.0') >= 0 ||
       substr(PHP_OS, 0, 3) !== 'WIN'))
   {
      $str = mcrypt_create_iv($len, MCRYPT_DEV_URANDOM);
      if ($str !== false)
         return $str;
   }


   /*
* No build-in crypto randomness function found. We collect any entropy
* available in the PHP core PRNGs along with some filesystem info and memory
* stats. To make this data cryptographically strong we add data either from
* /dev/urandom or if its unavailable, we gather entropy by measuring the
* time needed to compute a number of SHA-1 hashes.
*/
   $str = '';
   $bits_per_round = 2; // bits of entropy collected in each clock drift round
   $msec_per_round = 400; // expected running time of each round in microseconds
   $hash_len = 20; // SHA-1 Hash length
   $total = $len; // total bytes of entropy to collect

   $handle = @fopen('/dev/urandom', 'rb');
   if ($handle && function_exists('stream_set_read_buffer'))
      @stream_set_read_buffer($handle, 0);

   do
   {
      $bytes = ($total > $hash_len)? $hash_len : $total;
      $total -= $bytes;

      //collect any entropy available from the PHP system and filesystem
      $entropy = rand() . uniqid(mt_rand(), true) . $SSLstr;
      $entropy .= implode('', @fstat(@fopen( __FILE__, 'r')));
      $entropy .= memory_get_usage();
      if ($handle)
      {
         $entropy .= @fread($handle, $bytes);
      }
      else
      {
         // Measure the time that the operations will take on average
         for ($i = 0; $i < 3; $i ++)
         {
            $c1 = microtime(true);
            $var = sha1(mt_rand());
            for ($j = 0; $j < 50; $j++)
            {
               $var = sha1($var);
            }
            $c2 = microtime(true);
     $entropy .= $c1 . $c2;
         }

         // Based on the above measurement determine the total rounds
         // in order to bound the total running time.
         $rounds = (int)($msec_per_round*50 / (int)(($c2-$c1)*1000000));

         // Take the additional measurements. On average we can expect
         // at least $bits_per_round bits of entropy from each measurement.
         $iter = $bytes*(int)(ceil(8 / $bits_per_round));
         for ($i = 0; $i < $iter; $i ++)
         {
            $c1 = microtime();
            $var = sha1(mt_rand());
            for ($j = 0; $j < $rounds; $j++)
            {
               $var = sha1($var);
            }
            $c2 = microtime();
            $entropy .= $c1 . $c2;
         }

      }
      // We assume sha1 is a deterministic extractor for the $entropy variable.
      $str .= sha1($entropy, true);
   } while ($len > strlen($str));

   if ($handle)
      @fclose($handle);

   return substr($str, 0, $len);
}

/**
 * @param string $domain Pass $_SERVER['SERVER_NAME'] here
 * @param bool $debug
 *
 * @debug bool $debug
 * @return string
 */
function get_domain($domain, $debug = false)
{
	$original = $domain = strtolower($domain);

	if (filter_var($domain, FILTER_VALIDATE_IP)) { return $domain; }

	$debug ? print('<strong style="color:green">&raquo;</strong> Parsing: '.$original) : false;

	$arr = array_slice(array_filter(explode('.', $domain, 4), function($value){
		return $value !== 'www';
	}), 0); //rebuild array indexes

	if (count($arr) > 2)
	{
		$count = count($arr);
		$_sub = explode('.', $count === 4 ? $arr[3] : $arr[2]);

		$debug ? print(" (parts count: {$count})") : false;

		if (count($_sub) === 2) // two level TLD
		{
			$removed = array_shift($arr);
			if ($count === 4) // got a subdomain acting as a domain
			{
				$removed = array_shift($arr);
			}
			$debug ? print("<br>\n" . '[*] Two level TLD: <strong>' . join('.', $_sub) . '</strong> ') : false;
		}
		elseif (count($_sub) === 1) // one level TLD
		{
			$removed = array_shift($arr); //remove the subdomain

			if (strlen($_sub[0]) === 2 && $count === 3) // TLD domain must be 2 letters
			{
				array_unshift($arr, $removed);
			}
			else
			{
				// non country TLD according to IANA
				$tlds = array(
					'aero',
					'arpa',
					'asia',
					'biz',
					'cat',
					'com',
					'coop',
					'edu',
					'gov',
					'info',
					'jobs',
					'mil',
					'mobi',
					'museum',
					'name',
					'net',
					'org',
					'post',
					'pro',
					'tel',
					'travel',
					'xxx',
				);

				if (count($arr) > 2 && in_array($_sub[0], $tlds) !== false) //special TLD don't have a country
				{
					array_shift($arr);
				}
			}
			$debug ? print("<br>\n" .'[*] One level TLD: <strong>'.join('.', $_sub).'</strong> ') : false;
		}
		else // more than 3 levels, something is wrong
		{
			for ($i = count($_sub); $i > 1; $i--)
			{
				$removed = array_shift($arr);
			}
			$debug ? print("<br>\n" . '[*] Three level TLD: <strong>' . join('.', $_sub) . '</strong> ') : false;
		}
	}
	elseif (count($arr) === 2)
	{
		$arr0 = array_shift($arr);

		if (strpos(join('.', $arr), '.') === false
			&& in_array($arr[0], array('localhost','test','invalid')) === false) // not a reserved domain
		{
			$debug ? print("<br>\n" .'Seems invalid domain: <strong>'.join('.', $arr).'</strong> re-adding: <strong>'.$arr0.'</strong> ') : false;
			// seems invalid domain, restore it
			array_unshift($arr, $arr0);
		}
	}

	$debug ? print("<br>\n".'<strong style="color:gray">&laquo;</strong> Done parsing: <span style="color:red">' . $original . '</span> as <span style="color:blue">'. join('.', $arr) ."</span><br>\n") : false;

	return join('.', $arr);
}

# LOCK
//returns boolean: true=proceed, false=lock up; the difference between this and lockAction is that this can be used for repeated tasks, like admin
// then, only if action was unsuccessful would lockAction be called
function checkLock($action) {
	global $config, $db;
	$lock_time_initial = $config['lock_time_initial'];
	$lock_time_overload = $config['lock_time_overload'];
	$lock_count_overload = $config['lock_count_overload'];
	$lock_time_reset = $config['lock_time_reset'];
	$lock_time_max = $config['lock_time_max'];

	if(!isset($lock_time_initial[$action])) {
		return true; //well we can't do anything...
	}

	$ip = escape($_SERVER['REMOTE_ADDR']);
	$action = escape($action);

	$result = $db->query("SELECT id,time,num FROM locks WHERE ip='" . $ip . "' AND action='" . $action . "'");
	$row = $result->fetch_array();
	$result->close();

	if($row) {
		$id = $row['id'];
		$time = $row['time'];
		$count = $row['num']; //>=0 count means it's a regular initial lock; -1 count means overload lock

		if($count >= 0) {
			if(time() <= $time + $lock_time_initial[$action]) {
				return false;
			}
		} else {
			if(time() <= $time + $lock_time_overload[$action]) {
				return false;
			}
		}
	}

	return true;
}

//returns boolean: true=proceed, false=lock up
function lockAction($action) {
	global $config, $db;
	$lock_time_initial = $config['lock_time_initial'];
	$lock_time_overload = $config['lock_time_overload'];
	$lock_count_overload = $config['lock_count_overload'];
	$lock_time_reset = $config['lock_time_reset'];
	$lock_time_max = $config['lock_time_max'];

	if(!isset($lock_time_initial[$action])) {
		return true; //well we can't do anything...
	}

	$ip = escape($_SERVER['REMOTE_ADDR']);
	$action = escape($action);
	$replace_id = -1;

	//first find records with ip/action
	$result = $db->query("SELECT id,time,num FROM locks WHERE ip='" . $ip . "' AND action='" . $action . "'");
	$row = $result->fetch_array();

	if($row) {
		$id = $row['id'];
		$time = $row['time'];
		$count = $row['num']; //>=0 count means it's a regular initial lock; -1 count means overload lock

		if($count >= 0) {
			if(time() <= $time + $lock_time_initial[$action]) {
				return false;
			} else if(time() > $time + $lock_time_reset) {
				//this entry is old, but use it to replace
				$replace_id = $id;
			} else {
				//increase the count; maybe initiate an OVERLOAD
				$count = $count + 1;
				if($count >= $lock_count_overload[$action]) {
					$db->query("UPDATE locks SET num='-1', time='" . time() . "' WHERE ip='" . $ip . "'");
					return false;
				} else {
					$db->query("UPDATE locks SET num='" . $count . "', time='" . time() . "' WHERE ip='" . $ip . "'");
				}
			}
		} else {
			if(time() <= $time + $lock_time_overload[$action]) {
				return false;
			} else {
				//their overload is over, so this entry is old
				$replace_id = $id;
			}
		}
	} else {
		$db->query("INSERT INTO locks (ip, time, action, num) VALUES ('" . $ip . "', '" . time() . "', '" . $action . "', '1')");
	}

	if($replace_id != -1) {
		$db->query("UPDATE locks SET num='1', time='" . time() .  "' WHERE id='" . $replace_id . "'");
	}

	//some housekeeping
	$delete_time = time() - $lock_time_max;
	$db->query("DELETE FROM locks WHERE time<='" . $delete_time . "'");

	return true;
}

function execBackground($command) {
	return exec('php ' . escapeshellarg(includePath() . '/exec.php') . ' ' . escapeshellarg($command));
}

//returns false if failed to read log, or array of lines on success
function serviceGetLog($service_id, $type, $filename, $numlines = 400) {
	global $config;

	//get the identifier
	$id = stripAlphaNumeric(getServiceParam($service_id, "id"));

	if($id === false) {
		return "Error: the identifier for this service is not set.";
	}

	$log_file = $config["{$type}_path"] . $id . '/' . $filename;

	$jail = jailEnabled($service_id);

	if(($jail && !jailFileExists($service_id, $filename)) || (!$jail && !file_exists($log_file))) {
		return false;
	}

	//read last lines of the log file
	$output_array = array();

	if($jail) {
		jailExecute($service_id, "tail -n " . escapeshellarg($numlines) .  " " . escapeshellarg(jailPath($service_id) . $filename), $output_array);
	} else {
		exec("tail -n " . escapeshellarg($numlines) .  " " . escapeshellarg($log_file), $output_array);
	}

	return $output_array;
}

//gets the log quickly for AJAX updates
//database calls are eliminated by caching the "id" service parameter
// (requires reload if id changes)
//file is loaded quickly using tail + grep combination, based on the last line received
function serviceGetLogFast($service_id, $last_line, $type, $filename) {
	global $config;

	//sanity check on input
	if(strlen($last_line) > 2048) {
		return false;
	}

	//get the identifier
	if(!isset($_SESSION[$service_id . '_getlogfast_id'])) {
		$_SESSION[$service_id . '_getlogfast_id'] = stripAlphaNumeric(getServiceParam($service_id, "id"));
	}

	$id = $_SESSION[$service_id . '_getlogfast_id'];
	$log_file = $config["{$type}_path"] . $id . '/' . $filename;
	$jail = jailEnabled($service_id);

	if(($jail && !jailFileExists($service_id, $filename)) || (!$jail && !file_exists($log_file))) {
		return false;
	}

	//read last lines of the log file that client hasn't received yet
	$output_array = array();

	if($jail) {
		jailExecute($service_id, "tail -n 1000 " . escapeshellarg(jailPath($service_id) . $filename) . " | tac | fgrep -B 1000 -m 1 " . escapeshellarg($last_line) . " | tac", $output_array);
	} else {
		exec("tail -n 1000 " . escapeshellarg($log_file) . " | tac | fgrep -B 1000 -m 1 " . escapeshellarg($last_line) . " | tac", $output_array);
	}

	if(count($output_array) <= 1) {
		return false;
	} else {
		array_shift($output_array);
		return $output_array;
	}
}

function uxRedirect($url, $statusCode = 303) {
	header('Location: ' . $url, true, $statusCode);
	exit;
}

function httpRedirect($url) {
	echo '<html>
<head>
<meta http-equiv="Refresh" content="0; url=' . $url . '" />
</head>
<body>
<p>Please follow <a href="' . $url . '">this link</a>.</p>
</body>
</html>';
}

$UXPANEL_LOCK_FILENAME = NULL;
$UXPANEL_LOCK_FH = NULL;

function uxLock($name) {
	global $UXPANEL_LOCK_FH, $UXPANEL_LOCK_FILENAME;

	uxLockRelease();
	$UXPANEL_LOCK_FILENAME = '/var/lock/' . md5($name) . '.pid';
	$UXPANEL_LOCK_FH = @fopen($UXPANEL_LOCK_FILENAME, 'a');

	if(!$UXPANEL_LOCK_FH || !flock($UXPANEL_LOCK_FH, LOCK_EX | LOCK_NB, $eWouldBlock) || $eWouldBlock) {
		die('Failed to acquire lock.');
	} else {
		register_shutdown_function('uxLockRelease');
	}
}

function uxLockRelease() {
	global $UXPANEL_LOCK_FH, $UXPANEL_LOCK_FILENAME;

	if($UXPANEL_LOCK_FH !== NULL && $UXPANEL_LOCK_FILENAME !== NULL) {
		fclose($UXPANEL_LOCK_FH);
		unlink($UXPANEL_LOCK_FILENAME);

		$UXPANEL_LOCK_FH = NULL;
		$UXPANEL_LOCK_FILENAMK = NULL;
	}
}

?>
