<?php

if(php_sapi_name() !== 'cli') {
	die("Access forbidden.");
}

include("include/common.php");

uxLock('cronBilling');

include("config.php");
include("include/dbconnect.php");
include("include/account.php");
include("include/database.php");

$userInvoices = array(); // map from user ID to (map from time to services)
$result = $db->query("SELECT id, type, name FROM services");

while($row = $result->fetch_array()) {
	$service_id = $row[0];
	$service_type = $row[1];
	$service_name = $row[2];

	// load parameters
	$params = array();
	$result2 = $db->query("SELECT k, v FROM service_params WHERE service_id = '$service_id'");
	while($row2 = $result2->fetch_array()) {
		$params[$row2[0]] = $row2[1];
	}

	// skip if no due date or price
	// also parse price into amount and billing duration
	if(!isset($params['due']) || !isset($params['price']) || $params['due'] == 'N/A') {
		continue;
	}
	$priceParts = explode('$', $params['price']);
	if(count($priceParts) != 2) {
		continue;
	}
	$priceParts = explode(' ', $priceParts[1]);
	if(count($priceParts) != 2) {
		continue;
	}

	$priceMonths = false;
	if($priceParts[1] == 'monthly') {
		$priceMonths = 1;
	} else if($priceParts[1] == 'quarterly') {
		$priceMonths = 3;
	} else if($priceParts[1] == 'semi-annually') {
		$priceMonths = 6;
	} else if($priceParts[1] == 'annually') {
		$priceMonths = 12;
	} else {
		continue;
	}
	$amount = floatval($priceParts[0]);
	if($amount == 0) {
		continue;
	}

	echo "service $service_id ($service_type/$service_name) is due {$params['due']}\n";

	// check for invoice actions
	$dueTime = strtotime($params['due']);
	$invoiceStatus = isset($params['invoice_status']) ? $params['invoice_status'] : 0;

	if(($dueTime < time() + 7 * 24 * 3600 && $invoiceStatus < 1) || ($dueTime < time() + 1 * 24 * 3600 && $invoiceStatus < 2) || ($dueTime < time() - 5 * 24 * 3600 && $invoiceStatus < 3) || ($dueTime < time() - 10 * 24 * 3600 && $invoiceStatus < 4) || ($dueTime < time() - 15 * 24 * 3600 && $invoiceStatus < 5)) {
		echo "... action triggered!\n";
		$result2 = $db->query("SELECT email FROM accounts WHERE id = (SELECT MIN(account_id) FROM service_accounts WHERE service_id = '$service_id')");
		$row2 = $result2->fetch_array();
		if(!$row2) {
			continue;
		}
		$email = strtolower($row2[0]);
		if(!isset($userInvoices[$email])) {
			$userInvoices[$email] = array();
		}
		if(!isset($userInvoices[$email][$dueTime])) {
			$userInvoices[$email][$dueTime] = array();
		}
		$priceMonthsNice = $priceMonths . ($priceMonths > 1 ? ' months' : ' month');
		$userInvoices[$email][$dueTime][] = array('name' => $service_name, 'type' => $service_type, 'duration' => $priceMonthsNice, 'amount' => $amount);

		// update invoice status
		setServiceParam($service_id, 'invoice_status', $invoiceStatus + 1);
	} else if($invoiceStatus > 0 && $dueTime > time() + 8 * 24 * 3600) {
		// reset invoice status
		setServiceParam($service_id, 'invoice_status', 0);
	}
}

foreach($userInvoices as $email => $timeToService) {
	foreach($timeToService as $time => $services) {
		if($time < time() - 15 * 24 * 3600) {
			$subject = "Luna Ghost Service Suspension";
			$body = "We still have not received payment for the invoice below. As a result, services have been suspended. Please pay within two days to avoid service termination.";
		} else if($time < time() - 10 * 24 * 3600) {
			$subject = "Luna Ghost Overdue Invoice";
			$body = "The invoice below is overdue. If you wish to cancel any services, please e-mail support@lunaghost.com. Otherwise, please make a payment shortly to avoid service suspension.";
		} else if($time < time() - 5 * 24 * 3600) {
			$subject = "Luna Ghost Overdue Invoice";
			$body = "We have not received payment for this invoice. If you wish to cancel any services, please e-mail support@lunaghost.com. Otherwise, please make a payment soon.";
		} else if($time < time() + 1 * 24 * 3600) {
			$subject = "Luna Ghost Invoice Reminder";
			$body = "This is a reminder that the invoice below is due soon. Click the payment link below to pay via Paypal or Bitcoin. If you wish to cancel any services, please e-mail support@lunaghost.com.";
		} else {
			$subject = "Luna Ghost Invoice";
			$body = "A new invoice has been generated. Click the payment link below to pay via Paypal or Bitcoin.";
		}
		$dueNice = date("F j, Y", $time);
		$body .= "\n\nLuna Ghost Invoice (due $dueNice).\n";
		$totalDue = 0;
		foreach($services as $service) {
			$servicePrice = '$' . number_format($service['amount'], 2);
			$body .= "{$service['name']}: $servicePrice ({$service['duration']})\n";
			$totalDue += $service['amount'];
		}
		$totalDueNice = '$' . number_format($totalDue, 2);
		$body .= "Total due: $totalDueNice\n";
		$body .= "Payment link: https://lunaghost.com/paylg.php?account=$email&amount=$totalDue\n";
		$body .= "Due date: $dueNice\n\nBest,\nLuna Ghost";
		ux_mail($subject, $body, $email);
	}
}

?>
